﻿
using System.Collections.Generic;
using System.IO;
using System.Xml;
namespace HeuristicLab.Problems.RoutePlanning.Osm.Data {
  public class XmlDataSource : IDataSource {
    private FileInfo file;

    private IDictionary<long, Node> nodes;
    private IDictionary<long, Way> ways;
    private IDictionary<long, Relation> relations;

    private IDictionary<long, List<long>> nodeWays;


    #region Properties

    public bool IsReadOnly {
      get {
        if (file.Exists && file.IsReadOnly) {
          return true;
        }
        return false;
      }
    }

    public string Name {
      get { return file.Name; }
    }

    #endregion

    #region Constructors

    public XmlDataSource(string filename) {
      file = new FileInfo(filename);

      nodes = new Dictionary<long, Node>();
      ways = new Dictionary<long, Way>();
      relations = new Dictionary<long, Relation>();

      nodeWays = new Dictionary<long, List<long>>();

      ReadData();
    }

    #endregion

    #region Public Methods

    public Node GetNode(long nodeId) {
      if (nodes.ContainsKey(nodeId)) {
        return nodes[nodeId];
      }
      return null;
    }

    public Way GetWay(long wayId) {
      if (ways.ContainsKey(wayId)) {
        return ways[wayId];
      }
      return null;
    }

    public List<Way> GetWays(Node node) {
      if (nodeWays.ContainsKey(node.Id)) {
        List<long> wayIds = nodeWays[node.Id];
        List<Way> result = new List<Way>(wayIds.Count);
        for (int i = 0; i < wayIds.Count; i++) {
          result.Add(GetWay(wayIds[i]));
        }
        return result;
      }
      return null;
    }

    public Relation GetRelation(long relationId) {
      if (relations.ContainsKey(relationId)) {
        return relations[relationId];
      }
      return null;
    }

    #endregion

    private void ReadData() {
      XmlReaderSettings settings = new XmlReaderSettings();
      settings.IgnoreComments = true;
      settings.IgnoreWhitespace = true;

      using (XmlReader reader = XmlReader.Create(file.FullName, settings)) {
        reader.ReadStartElement("osm");
        reader.ReadStartElement("bounds");
        while (reader.LocalName.Equals("node")) {
          Node node = new Node();

          double lat = XmlConvert.ToDouble(reader.GetAttribute("lat"));
          double lon = XmlConvert.ToDouble(reader.GetAttribute("lon"));

          GetBaseAttributes(reader, node);
          node.Latitude = lat;
          node.Longitude = lon;

          reader.Read();
          if (reader.LocalName.Equals("tag")) {
            while (reader.LocalName.Equals("tag")) {
              string key = reader.GetAttribute("k");
              string value = reader.GetAttribute("v");
              node.Tags.Add(key, value);
              reader.Read();
            }
            reader.Read();
          }

          nodes.Add(node.Id, node);
        }
        while (reader.LocalName.Equals("way")) {
          Way way = new Way();
          GetBaseAttributes(reader, way);
          reader.Read();
          while (reader.LocalName.Equals("nd") || reader.LocalName.Equals("tag")) {
            if (reader.LocalName.Equals("nd")) {
              long refNodeId = XmlConvert.ToInt64(reader.GetAttribute("ref"));
              way.Nodes.Add(GetNode(refNodeId));

            } else if (reader.LocalName.Equals("tag")) {
              string key = reader.GetAttribute("k");
              string value = reader.GetAttribute("v");
              way.Tags.Add(key, value);
            }
            reader.Read();
          }

          ways.Add(way.Id, way);
          InsertNodeWayRelation(way);

          reader.Read();
        }
        while (reader.LocalName.Equals("relation")) {
          Relation relation = new Relation();
          GetBaseAttributes(reader, relation);

          reader.Read();
          while (reader.LocalName.Equals("member") || reader.LocalName.Equals("tag")) {
            if (reader.LocalName.Equals("member")) {
              string type = reader.GetAttribute("type");
              long refNodeId = XmlConvert.ToInt64(reader.GetAttribute("ref"));
              string role = reader.GetAttribute("role");
              RelationMember member = new RelationMember();
              member.Role = role;
              if (type == "node") {
                member.Member = GetNode(refNodeId);
              } else if (type == "way") {
                member.Member = GetWay(refNodeId);
              }
              relation.Members.Add(member);
            } else if (reader.LocalName.Equals("tag")) {
              string key = reader.GetAttribute("k");
              string value = reader.GetAttribute("v");
              relation.Tags.Add(key, value);
            }
            reader.Read();
          }

          relations.Add(relation.Id, relation);

          reader.Read();
        }
        reader.Close();
      }
    }

    private void GetBaseAttributes(XmlReader reader, OsmBase obj) {
      obj.Id = XmlConvert.ToInt64(reader.GetAttribute("id"));
      //Console.WriteLine("id:" + obj.Id);

      if (reader.GetAttribute("user") != null) {
        obj.User = reader.GetAttribute("user");
      }
      if (reader.GetAttribute("uid") != null) {
        obj.UserId = XmlConvert.ToInt32(reader.GetAttribute("uid"));
      }
      obj.Visible = XmlConvert.ToBoolean(reader.GetAttribute("visible"));
      obj.Version = XmlConvert.ToInt32(reader.GetAttribute("version"));
      obj.Changeset = XmlConvert.ToInt64(reader.GetAttribute("changeset"));
      obj.Timestamp = XmlConvert.ToDateTime(reader.GetAttribute("timestamp"), XmlDateTimeSerializationMode.Utc);

    }

    private void InsertNodeWayRelation(Way way) {
      foreach (Node node in way.Nodes) {
        if (!nodeWays.ContainsKey(node.Id)) {
          nodeWays.Add(node.Id, new List<long>());
        }
        nodeWays[node.Id].Add(way.Id);
      }
    }
  }
}
